#version 450

layout (location = 0) in vec3 fragNormal;      
layout (location = 1) in vec3 fragPosition;    
layout (location = 2) in vec4 originalColor;   

layout (binding = 0) uniform UniformBuffer {
    vec4 color;
    mat4 modelViewProjMatrix;
    mat4 modelMatrix;
    vec4 cameraPosition;
    ivec4 userInput;
    float ambientConstant;
    float diffuseConstant;
    float specularConstant;
    float shininessConstant;
} ub_data;

layout (binding = 1) uniform DirectionalLightBuffer {
    vec4 direction;
    vec4 color;
} dir_data;

layout (binding = 2) uniform PointLightBuffer {
    vec4 position;
    vec4 color;
    vec4 attenuation;
} point_data;

layout (location = 0) out vec4 out_color;

void main() {

    vec3 normalizedNormal = normalize(fragNormal);
    vec3 viewDir = normalize(ub_data.cameraPosition.xyz - fragPosition);
    
    
    if (ub_data.userInput[0] == 1) {
        if (dot(normalizedNormal, viewDir) > 0.0) {
            out_color = vec4(fragNormal, 1.0f); // Normal view
        } else {
            out_color = vec4(-fragNormal, 1.0f);
        }
    } else {
        
        if (dot(normalizedNormal, viewDir) > 0.0) {
            // Emissive Illumination - front face
            out_color = originalColor;
        } else {
            
            //Back face - Phong 
            normalizedNormal = -normalizedNormal; // Invert normal

            // Ambient component
            vec3 ambient = originalColor.rgb;

            // Diffuse component Directional light
            vec3 lightDir = normalize(dir_data.direction.xyz);
            float diffDir = max(dot(normalizedNormal, lightDir), 0.0);
            vec3 diffuseDir = diffDir * originalColor.rgb;

            // Diffuse component Point light
            vec3 lightDirPoint = normalize(point_data.position.xyz - fragPosition);
            float diffPoint = max(dot(normalizedNormal, lightDirPoint), 0.0);
            vec3 diffusePoint = diffPoint * originalColor.rgb;

            // Attenuation
            float distance = length(point_data.position.xyz - fragPosition);
            float attenuationFactor = 1.0 / (point_data.attenuation.x + point_data.attenuation.y * distance + point_data.attenuation.z * distance * distance);
            diffusePoint *= attenuationFactor;

            // Specular component
            float shininess = ub_data.shininessConstant;
            vec3 specular = vec3(0.0);

            // Directional Light Specular
            vec3 reflectDirDir = reflect(-normalize(dir_data.direction.xyz), normalizedNormal);
            float specDir = pow(max(dot(viewDir, reflectDirDir), 0.0), shininess);
            specular += specDir * originalColor.rgb;

            // Point Light Specular
            vec3 lightDirPointSpec = normalize(point_data.position.xyz - fragPosition);
            vec3 reflectDirPointSpec = reflect(-lightDirPoint, normalizedNormal);
            float specPoint = pow(max(dot(viewDir, reflectDirPointSpec), 0.0), shininess);
            specular += specPoint * originalColor.rgb;
            
            vec3 resultColor = ub_data.ambientConstant * ambient + ub_data.diffuseConstant *  (diffusePoint + diffuseDir) + ub_data.specularConstant * specular;

            // Light color
            resultColor *= (dir_data.color.rgb + point_data.color.rgb);

            out_color = vec4(resultColor, 1.0);
        }
    }
}
