#version 450

layout(location = 0) in VertexData {
    vec3 fragPositionWS;
    vec3 fragNormalWS;
    vec2 textureCoordinates;
    vec3 fragTangentWS;
    vec3 fragBitangentWS;
    vec4 fragPositionLightSpaceCS;
} frag_in;

layout(binding = 1) uniform sampler2D diffuse_texture;
layout(binding = 2) uniform sampler2D normal_texture;
layout(binding = 3) uniform sampler2D height_texture;
layout(binding = 4) uniform sampler2D imageShadowMap;

layout(binding = 0) uniform UniformBuffer {
    mat4 viewProjMatrixShadow;
    vec4 spotlightPositionWS; //w = farPlane
	vec4 spotlightDirectionWS;
    vec4 color;
    mat4 modelViewProjMatrix;
    mat4 modelMatrix;
    vec4 cameraPosition;
    ivec4 userInput;
    float ambientConstant;
    float diffuseConstant;
    float specularConstant;
    float shininessConstant;
} ub_data;

layout(location = 0) out vec4 out_color;

vec2 parallaxMapping(vec2 texCoords, vec3 viewDir) {
    float heightScale = 0.1; // Adjust this value as needed
    float height = texture(height_texture, texCoords).r;
    return texCoords + (height * heightScale * viewDir.xy);
}


float calculateShadowFactor(vec3 fragPosition, vec3 lightPosition)
{
		// vector from light source to fragment position yields distance from light to fragment (occludee)
        vec4 fragPosLightSpace_CS = (ub_data.viewProjMatrixShadow * vec4(fragPosition,1));
        vec4 fragPosLightSpace_NDC = (fragPosLightSpace_CS/fragPosLightSpace_CS.w);
        fragPosLightSpace_NDC.xy = 0.5 * fragPosLightSpace_NDC.xy + 0.5; 
        float normalizedDistanceToOccludingFrag = texture(imageShadowMap, fragPosLightSpace_NDC.xy).r;
        float shadowFactor = 0;
        float bias = 0.00007;
        float normalizedDistanceFragToLight = fragPosLightSpace_NDC.z;

        // Average shadow in neighbourhood (pcf)
        vec2 texelSize = 1.0 / textureSize(imageShadowMap, 0);
        for(int x = -1; x <= 1; ++x)
        {
        for(int y = -1; y <= 1; ++y)
            {
                float normalizedDistanceSampleToLight = texture(imageShadowMap, fragPosLightSpace_NDC.xy + vec2(x, y) * texelSize).r; 
                shadowFactor += normalizedDistanceSampleToLight < normalizedDistanceFragToLight - bias  ? 0.0 : 1.0;        
            }    
        }
        shadowFactor /= 9;
        return shadowFactor;
}

float calculateSpotlightIntensity(vec3 lightPosition, vec3 lightDirection, vec3 fragPos)
{
		float spotlightIntensity = 0;
		
        // parameters for spotlight 
        float cutOff = 0.97629;//cos(radians(12.5f)));
        float outerCutOff = 0.95371695074;; //cos(radians(17.5f)));
		float scalarProduct = max(0, dot(normalize(lightDirection), normalize(fragPos-lightPosition)));
		
        // if fragment in light cone caclulate light intensity inlcuding shadow
        if( scalarProduct > outerCutOff )  
		{      
            // parameters for spotlight
            float linAttenFact = 0.09;
            float quadrAttenFact = 0.032;

            //Attenuation for spotLight
		    float distanceSpotlight = length(lightPosition - fragPos);
		    float attenuationSpotlight = 1.0 / (1 + linAttenFact * distanceSpotlight + quadrAttenFact *(distanceSpotlight*distanceSpotlight));
            float coneIntensity = (scalarProduct - outerCutOff) / (cutOff-outerCutOff);
			coneIntensity = clamp(coneIntensity, 0.0, 1.0);
            float shadowFactor =  calculateShadowFactor(fragPos,lightPosition);
            spotlightIntensity = shadowFactor * coneIntensity * attenuationSpotlight;
		}
		return spotlightIntensity;
	}


void main() {
    vec3 viewDir = normalize(ub_data.cameraPosition.xyz - frag_in.fragPositionWS);
    vec3 T = normalize(frag_in.fragTangentWS);
    vec3 B = normalize(frag_in.fragBitangentWS);
    vec3 N = normalize(frag_in.fragNormalWS);
    mat3 TBN = mat3(T, B, N);
    viewDir = TBN * viewDir;

    // Parallax mapping to adjust texture coordinates
    vec2 texCoords = parallaxMapping(frag_in.textureCoordinates, viewDir);

    // Debugging output for texture coordinates
    // out_color = vec4(texCoords, 0.0, 1.0);
    
    // Sample the textures with adjusted coordinates
    vec3 diffuseColor = texture(diffuse_texture, texCoords).rgb;
    vec3 normalMap = texture(normal_texture, texCoords).rgb;

    // Debugging output for texture colors
    // out_color = vec4(diffuseColor, 1.0);
    // return;

    // Transform the normal map value from [0, 1] to [-1, 1]
    vec3 normalTangentSpace = normalize(normalMap * 2.0 - 1.0);

    // Transform the normal from tangent space to world space
    vec3 normalWS = normalize(TBN * normalTangentSpace);

    // Debugging output for normals
    // out_color = vec4(normalWS * 0.5 + 0.5, 1.0);
    // return;

    // Lighting calculations
    vec3 lightPos = vec3(10.0, 10.0, 10.0);
    vec3 lightColor = 0.1 * vec3(1.0, 1.0, 1.0);

    vec3 viewPos = vec3(ub_data.cameraPosition);
    vec3 lightDir = normalize(lightPos - frag_in.fragPositionWS);

    // Ambient component
    vec3 ambient = ub_data.ambientConstant * diffuseColor;

    // Diffuse component
    float diff = max(dot(normalWS, lightDir), 0.0);
    vec3 diffuse = ub_data.diffuseConstant * diff * diffuseColor;

    // Specular component
    vec3 reflectDir = reflect(-lightDir, normalWS);
    float spec = pow(max(dot(viewDir, reflectDir), 0.0), ub_data.shininessConstant);
    vec3 specular = ub_data.specularConstant * spec * vec3(1.0);

     // Calculate diffuse spotlight with shadowFactor
    vec3 spotlightDir = normalize(ub_data.spotlightDirectionWS.xyz);
    vec3 spotlightColor = vec3(1,1,1);
    float spotlightIntensity = calculateSpotlightIntensity(ub_data.spotlightPositionWS.xyz, ub_data.spotlightDirectionWS.xyz, frag_in.fragPositionWS);
    float diffSpotlight = 0.2;//*max(dot(normalWS, -spotlightDir), 0.0);
    vec3 diffuseSpotlight = ub_data.diffuseConstant * diffSpotlight * diffuseColor * spotlightIntensity * spotlightColor;
    
    // Calculate specular spotlight with shadowFactor
    vec3 reflectDirSpotlight = reflect(-spotlightDir, normalWS);
    float specSpotlight = pow(max(dot(viewDir, reflectDir), 0.0), ub_data.shininessConstant);
    vec3 specularSpotlight = ub_data.specularConstant * specSpotlight * vec3(1.0) * spotlightIntensity * spotlightColor;

    // Combine results
    vec3 result = (ambient + diffuse + specular) * lightColor + diffuseSpotlight + specularSpotlight;
    out_color = vec4(result, 1.0);
}
