#version 450

layout (location = 0) in vec3 fragNormal;      // Normal in world space
layout (location = 1) in vec3 fragPosition;    // Position in world space
layout (location = 2) in vec4 originalColor;   // Original color

layout (binding = 0) uniform UniformBuffer {
    vec4 color;
    mat4 modelViewProjMatrix;
    mat4 modelMatrix;
    vec4 cameraPosition;
    ivec4 userInput;
    float ambientConstant;
    float diffuseConstant;
    float specularConstant;
    float shininessConstant;
} ub_data;

layout (binding = 1) uniform DirectionalLightBuffer {
    vec4 direction;
    vec4 color;
} dir_data;

layout (binding = 2) uniform PointLightBuffer {
    vec4 position;
    vec4 color;
    vec4 attenuation;
} point_data;

layout (location = 0) out vec4 out_color;

// Gets the reflected color value from a a certain position, from
// a certain direction INSIDE of a cornell box of size 3 which is
// positioned at the origin.
// positionWS:  Position inside the cornell box for which to get the
//              reflected color value for from -directionWS.
// directionWS: Outgoing direction vector (from positionWS towards the // outside) for which to get the reflected color value for.
vec3 getCornellBoxReflectionColor(vec3 positionWS, vec3 directionWS) {
    vec3 P0 = positionWS;
    vec3 V = normalize(directionWS);
    const float boxSize = 1.5; vec4[5] planes = {
    vec4(-1.0, 0.0, 0.0, -boxSize), // left
    vec4( 1.0, 0.0, 0.0, -boxSize), // right
    vec4( 0.0, 1.0, 0.0, -boxSize), // top
    vec4( 0.0, -1.0, 0.0, -boxSize), // bottom
    vec4( 0.0, 0.0, -1.0, -boxSize) // back
    };
    vec3[5] colors = {
    vec3(1.0, 0.0, 0.0), // left
    vec3(0.0, 1.0, 0.0), // right
    vec3(0.96, 0.93, 0.85), // top
    vec3(0.64, 0.64, 0.64), // bottom
    vec3(0.76, 0.74, 0.68) // back
    };
    for (int i = 0; i < 5; ++i) {
        vec3 N = planes[i].xyz;
        float d = planes[i].w;
        float denom = dot(V, N);
        if (denom <= 0) continue;
        float t = -(dot(P0, N) + d)/denom; vec3 P=P0+t*V;
        float q = boxSize + 0.01;
        if (P.x > -q && P.x < q && P.y > -q && P.y < q && P.z > -q && P.z < q) {
            return colors[i]; }
    }
    return vec3(0.0, 0.0, 0.0);
}

vec3 clampedReflect(vec3 I, vec3 N) {
    return I - 2.0 * min(dot(N, I), 0.0) * N;
}

void main() {
    if (ub_data.userInput[0] == 1) {
        out_color = vec4(fragNormal, 1.0); // Normal view
    } else {
        vec3 normalizedNormal = normalize(fragNormal);
        vec3 viewDir = normalize(ub_data.cameraPosition.xyz - fragPosition);

        // Reflection direction
        vec3 reflectDir = clampedReflect(-viewDir, normalize(fragNormal));

        // Reflection color
        vec3 reflectionColor = getCornellBoxReflectionColor(fragPosition, reflectDir);

        // Fresnel term (Schlick's approximation)
        float R0 = pow((2.0 - 1.0) / (2.0 + 1.0), 2.0); // IOR 1.5
        float cosTheta = dot(viewDir, normalize(fragNormal));
        float fresnel = R0 + (1.0 - R0) * pow(1.0 - cosTheta, 5.0);

        // Ambient component
        vec3 ambientFinalComponent = ub_data.ambientConstant * originalColor.rgb;

        // Directional light
        vec3 lightDir = normalize(dir_data.direction.xyz);
        float diffDir = max(dot(normalizedNormal, lightDir), 0.0);
        vec3 diffuseDir = diffDir * ub_data.diffuseConstant * dir_data.color.rgb;

        // Point light
        vec3 lightDirPoint = normalize(point_data.position.xyz - fragPosition);
        float diffPoint = max(dot(normalizedNormal, lightDirPoint), 0.0);
        vec3 diffusePoint = diffPoint * ub_data.diffuseConstant * point_data.color.rgb;

        // Attenuation
        float distance = length(point_data.position.xyz - fragPosition);
        float attenuationFactor = 1.0 / (point_data.attenuation.x + point_data.attenuation.y * distance + point_data.attenuation.z * distance * distance);
        diffusePoint *= attenuationFactor;
        \
        // Diffuse component
        vec3 diffuseFinalComponent = (diffuseDir + diffusePoint) * ub_data.diffuseConstant * originalColor.rgb;


        // Specular component
        float shininess = ub_data.shininessConstant;
        vec3 specular = vec3(0.0);

        // Directional Light Specular
        vec3 reflectDirDir = reflect(-normalize(dir_data.direction.xyz), normalizedNormal);
        float specDir = pow(max(dot(viewDir, reflectDirDir), 0.0), shininess);
        specular += specDir * dir_data.color.rgb;

        // Point Light Specular
        vec3 lightDirPointSpec = normalize(point_data.position.xyz - fragPosition);
        vec3 reflectDirPointSpec = reflect(-lightDirPoint, normalizedNormal);
        float specPoint = pow(max(dot(viewDir, reflectDirPointSpec), 0.0), shininess);
        specular += specPoint * point_data.color.rgb;

        // Specular component
        vec3 specularFinalComponent = specular * ub_data.specularConstant * originalColor.rgb;

        vec3 resultColor = ambientFinalComponent + diffuseFinalComponent + specularFinalComponent;

        vec3 finalColor = mix(resultColor, reflectionColor, fresnel);

        out_color = vec4(finalColor, 1.0);
    }
}
