#version 450

layout(location = 0) in VertexData {
    vec3 fragPositionWS;
    vec3 fragNormalWS;
    vec2 textureCoordinates;
    vec3 fragTangentWS;
    vec3 fragBitangentWS;
    vec4 fragPositionLightSpaceCS;
} frag_in;

layout(binding = 1) uniform sampler2D diffuse_texture;
layout(binding = 2) uniform sampler2D specular_texture;
layout(binding = 3) uniform sampler2D normal_texture;
layout(binding = 4) uniform sampler2D imageShadowMap;

layout(binding = 0) uniform UniformBuffer {
    mat4 viewProjMatrixShadow;
    vec4 spotlightPositionWS; //w = farPlane
	vec4 spotlightDirectionWS;
    vec4 color;
    mat4 modelViewProjMatrix;
    mat4 modelMatrix;
    vec4 cameraPosition;
    ivec4 userInput;
    float ambientConstant;
    float diffuseConstant;
    float specularConstant;
    float shininessConstant;
} ub_data;

layout(location = 0) out vec4 out_color;


float calculateShadowFactor(vec3 fragPosition, vec3 lightPosition, vec3 fragNormal, vec3 lightDir)
{
		// vector from light source to fragment position yields distance from light to fragment (occludee)
        vec4 fragPosLightSpace_CS = (ub_data.viewProjMatrixShadow * vec4(fragPosition,1));
        vec4 fragPosLightSpace_NDC = (fragPosLightSpace_CS/fragPosLightSpace_CS.w);
        fragPosLightSpace_NDC.xy = 0.5 * fragPosLightSpace_NDC.xy + 0.5; 
        float normalizedDistanceToOccludingFrag = texture(imageShadowMap, fragPosLightSpace_NDC.xy).r;
        float shadowFactor = 0;
        float minBias = 0.00001;
        float maxBias = 0.000019;
        float bias = max(maxBias * (1.0 - dot(fragNormal, lightDir)), minBias); 
        float normalizedDistanceFragToLight = fragPosLightSpace_NDC.z;

        // Average shadow in neighbourhood (pcf)
        vec2 texelSize = 1.0 / textureSize(imageShadowMap, 0);
        for(int x = -1; x <= 1; ++x)
        {
        for(int y = -1; y <= 1; ++y)
            {
                float normalizedDistanceSampleToLight = texture(imageShadowMap, fragPosLightSpace_NDC.xy + vec2(x, y) * texelSize).r; 
                shadowFactor += normalizedDistanceSampleToLight < normalizedDistanceFragToLight - bias  ? 0.0 : 1.0;        
            }    
        }
        shadowFactor /= 9;
        return shadowFactor;
}

float calculateSpotlightIntensity(vec3 lightPosition, vec3 fragPos,vec3 lightDirection, vec3 fragNormal)
{
		float spotlightIntensity = 0;
		
        // parameters for spotlight 
        float cutOff = 0.97629;//cos(radians(12.5f)));
        float outerCutOff = 0.95371695074;; //cos(radians(17.5f)));
		float scalarProduct = max(0, dot(normalize(lightDirection), normalize(fragPos-lightPosition)));
		
        // if fragment in light cone caclulate light intensity inlcuding shadow
        if( scalarProduct > outerCutOff )  
		{      
            // parameters for spotlight
            float linAttenFact = 0.045; // this value is for 100m reach use 0.09 for 50m reach
            float quadrAttenFact = 0.0075; // this value is for 100m reach use  0.032 for 50m reach

            //Attenuation for spotLight
		    float distanceSpotlight = length(lightPosition - fragPos);
		    float attenuationSpotlight = 1.0 / (1 + linAttenFact * distanceSpotlight + quadrAttenFact *(distanceSpotlight*distanceSpotlight));
            float coneIntensity = (scalarProduct - outerCutOff) / (cutOff-outerCutOff);
			coneIntensity = clamp(coneIntensity, 0.0, 1.0);
            float shadowFactor =  calculateShadowFactor(fragPos,lightPosition, fragNormal,lightDirection );
            spotlightIntensity = shadowFactor * coneIntensity * attenuationSpotlight;
		}
		return spotlightIntensity;
	}


void main() {
    // Sample the textures
    vec3 diffuseColor = texture(diffuse_texture, frag_in.textureCoordinates).rgb;
    vec3 specularColor = texture(specular_texture, frag_in.textureCoordinates).rgb;
    vec3 normalMap = texture(normal_texture, frag_in.textureCoordinates).rgb;

    // Transform the normal map value from [0, 1] to [-1, 1]
    vec3 normalTangentSpace = normalize(normalMap * 2.0 - 1.0);

    // Construct the TBN matrix
    mat3 TBN = mat3(frag_in.fragTangentWS, frag_in.fragBitangentWS, frag_in.fragNormalWS);

    // Transform the normal from tangent space to world space
    vec3 normalWS = normalize(TBN * normalTangentSpace);

    // Lighting calculations
    vec3 lightPos = vec3(10.0, 10.0, 10.0);
    vec3 lightColor = vec3(1.0, 1.0, 1.0);

    vec3 viewPos = vec3(ub_data.cameraPosition);
    vec3 viewDir = normalize(viewPos - frag_in.fragPositionWS);
    vec3 lightDir = normalize(lightPos - frag_in.fragPositionWS);

    // Ambient component
    vec3 ambient = ub_data.ambientConstant * diffuseColor;

    // Diffuse component
    float diff = max(dot(normalWS, lightDir), 0.0);
    vec3 diffuse = ub_data.diffuseConstant * diff * diffuseColor;

    // Specular component
    vec3 reflectDir = reflect(-lightDir, normalWS);
    float spec = pow(max(dot(viewDir, reflectDir), 0.0), ub_data.shininessConstant);
    vec3 specular = ub_data.specularConstant * spec * specularColor;

    // Calculate diffuse spotlight with shadowFactor
    vec3 spotlightDir = normalize(ub_data.spotlightPositionWS.xyz-frag_in.fragPositionWS); // ub_data.spotlightDirectionWS points from poinlight away!
    vec3 spotlightColor = vec3(1,1,1);
    float spotlightIntensity = calculateSpotlightIntensity(ub_data.spotlightPositionWS.xyz, frag_in.fragPositionWS, ub_data.spotlightDirectionWS.xyz,  frag_in.fragNormalWS);
    float diffSpotlight = max(dot(normalWS, spotlightDir), 0.0); 
    if(diffSpotlight <= 0) diffSpotlight = 0.2; // this is a hack to avoid problems with normal maps
    vec3 diffuseSpotlight = ub_data.diffuseConstant * diffSpotlight * diffuseColor * spotlightIntensity * spotlightColor;
    
    // Calculate specular spotlight with shadowFactor
    vec3 reflectDirSpotlight = reflect(-spotlightDir, normalWS);
    float specSpotlight = pow(max(dot(viewDir, reflectDir), 0.0), ub_data.shininessConstant);
    vec3 specularSpotlight = ub_data.specularConstant * specSpotlight * specularColor * spotlightIntensity * spotlightColor;
    
    // Combine results
    vec3 result = (ambient + diffuse + specular) * lightColor * 0.1+ (diffuseSpotlight + specularSpotlight);
    out_color = vec4(result, 1.0);
}
